import importlib
import os
from contextlib import contextmanager


def truncate_with_epsilons(value, max_length=32, epsilon="..."):
    if len(str(value)) > max_length:
        truncated_value = str(value)[: max_length - len(epsilon)] + epsilon
    else:
        truncated_value = str(value)
    return truncated_value


@contextmanager
def change_directory(new_path):
    # Save the current working directory
    current_path = os.getcwd()

    try:
        # Change the working directory
        os.chdir(new_path)
        yield
    finally:
        # Change back to the original working directory
        os.chdir(current_path)


def directly_import_module(name, path):
    """Returns the named Python module loaded from the path"""
    assert path.exists()
    spec = importlib.util.spec_from_file_location(name, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    return module



import json
import logging
import uuid
from copy import deepcopy
from importlib import metadata
from pathlib import Path

from grand_challenge_forge.generation_utils import (
    copy_and_render,
    generate_socket_value_stub_file,
    socket_to_socket_value,
)
from grand_challenge_forge.schemas import (
    validate_algorithm_template_context,
    validate_pack_context,
)

logger = logging.getLogger(__name__)


def generate_challenge_pack(
    *,
    output_zip_file,
    target_zpath,
    context,
):
    validate_pack_context(context)

    context["grand_challenge_forge_version"] = metadata.version(
        "grand-challenge-forge"
    )

    # Generate the README.md file
    copy_and_render(
        templates_dir_name="pack-readme",
        output_zip_file=output_zip_file,
        target_zpath=target_zpath,
        context=context,
    )

    for phase in context["challenge"]["phases"]:
        phase_zpath = target_zpath / phase["slug"]
        phase_context = {"phase": phase}

        generate_upload_to_archive_script(
            context=phase_context,
            output_zip_file=output_zip_file,
            target_zpath=phase_zpath
            / f"upload-to-archive-{phase['archive']['slug']}",
        )

        generate_example_algorithm(
            context=phase_context,
            output_zip_file=output_zip_file,
            target_zpath=phase_zpath / "example-algorithm",
        )

        generate_example_evaluation(
            context=phase_context,
            output_zip_file=output_zip_file,
            target_zpath=phase_zpath / "example-evaluation-method",
        )


def generate_upload_to_archive_script(
    *,
    output_zip_file,
    target_zpath,
    context,
):
    context = deepcopy(context)

    expected_cases_per_interface = {}
    for idx, interface in enumerate(context["phase"]["algorithm_interfaces"]):
        interface_name = f"interface_{idx}"
        archive_cases = generate_archive_cases(
            inputs=interface["inputs"],
            output_zip_file=output_zip_file,
            target_zpath=target_zpath / interface_name,
            number_of_cases=3,
        )

        # Make cases relative to the script
        for case in archive_cases:
            for k, v in case.items():
                case[k] = Path(*v.parts[1:])

        expected_cases_per_interface[interface_name] = archive_cases

    all_algorithm_inputs = {}
    for interface in context["phase"]["algorithm_interfaces"]:
        for socket in interface["inputs"]:
            all_algorithm_inputs[socket["slug"]] = socket

    context.update(
        {
            "all_algorithm_inputs": all_algorithm_inputs,
            "expected_cases_per_interface": expected_cases_per_interface,
        }
    )

    copy_and_render(
        templates_dir_name="upload-to-archive-script",
        output_zip_file=output_zip_file,
        target_zpath=target_zpath,
        context=context,
    )


def generate_archive_cases(
    *, inputs, output_zip_file, target_zpath, number_of_cases
):
    result = []
    for i in range(0, number_of_cases):
        item_files = {}
        for input_socket in inputs:
            # Use deep zpath to create the files
            zpath = (
                target_zpath
                / Path(f"case_{i}")
                / input_socket["relative_path"]
            )

            # Report back relative to script paths
            generate_socket_value_stub_file(
                output_zip_file=output_zip_file,
                target_zpath=zpath,
                socket=input_socket,
            )

            item_files[input_socket["slug"]] = zpath

        result.append(item_files)

    return result


def _interface_context(interfaces):
    # Build context
    algorithm_input_sockets = [
        socket for interface in interfaces for socket in interface["inputs"]
    ]
    algorithm_output_sockets = [
        socket for interface in interfaces for socket in interface["outputs"]
    ]

    algorithm_interface_keys = []
    for interface in interfaces:
        algorithm_interface_keys.append(
            tuple(sorted([socket["slug"] for socket in interface["inputs"]]))
        )

    interface_names = [f"interface_{idx}" for idx, _ in enumerate(interfaces)]

    return {
        "algorithm_interface_names": interface_names,
        "algorithm_interface_keys": algorithm_interface_keys,
        "algorithm_input_sockets": algorithm_input_sockets,
        "algorithm_output_sockets": algorithm_output_sockets,
    }


def generate_example_algorithm(*, output_zip_file, target_zpath, context):
    context = deepcopy(context)

    interface_names = []
    for idx, interface in enumerate(context["phase"]["algorithm_interfaces"]):
        interface_name = f"interface_{idx}"
        interface_names.append(interface_name)

        input_zdir = target_zpath / "test" / "input" / interface_name
        inputs = interface["inputs"]

        # create inputs.json
        output_zip_file.writestr(
            str(input_zdir / "inputs.json"),
            json.dumps(
                [socket_to_socket_value(socket) for socket in inputs], indent=4
            ),
        )

        # Create actual input files
        for input in inputs:
            generate_socket_value_stub_file(
                output_zip_file=output_zip_file,
                target_zpath=input_zdir / input["relative_path"],
                socket=input,
            )

    context.update(
        _interface_context(interfaces=context["phase"]["algorithm_interfaces"])
    )

    copy_and_render(
        templates_dir_name="example-algorithm",
        output_zip_file=output_zip_file,
        target_zpath=target_zpath,
        context=context,
    )


def generate_example_evaluation(*, output_zip_file, target_zpath, context):
    context = deepcopy(context)
    context.update(
        _interface_context(interfaces=context["phase"]["algorithm_interfaces"])
    )

    input_zdir = target_zpath / "test" / "input"

    predictions_json = []
    for interface in context["phase"]["algorithm_interfaces"]:
        predictions_json.extend(
            generate_predictions_json(
                inputs=interface["inputs"],
                outputs=interface["outputs"],
                number_of_jobs=3,
            )
        )

    output_zip_file.writestr(
        str(input_zdir / "predictions.json"),
        json.dumps(predictions_json, indent=4),
    )

    generate_prediction_files(
        output_zip_file=output_zip_file,
        target_zpath=target_zpath / "test" / "input",
        predictions=predictions_json,
    )

    for socket in context["phase"]["evaluation_additional_inputs"]:
        generate_socket_value_stub_file(
            output_zip_file=output_zip_file,
            target_zpath=input_zdir / socket["relative_path"],
            socket=socket,
        )

    copy_and_render(
        templates_dir_name="example-evaluation-method",
        output_zip_file=output_zip_file,
        target_zpath=target_zpath,
        context=context,
    )


def generate_predictions_json(
    *,
    inputs,
    outputs,
    number_of_jobs,
):
    predictions = []
    for _ in range(0, number_of_jobs):
        predictions.append(
            {
                "pk": str(uuid.uuid4()),
                "inputs": [
                    socket_to_socket_value(socket) for socket in inputs
                ],
                "outputs": [
                    socket_to_socket_value(socket) for socket in outputs
                ],
                "status": "Succeeded",
                "started_at": "2024-11-29T10:31:25.691799Z",
                "completed_at": "2024-11-29T10:31:50.691799Z",
            }
        )
    return predictions


def generate_prediction_files(*, output_zip_file, target_zpath, predictions):
    for prediction in predictions:
        prediction_zpath = target_zpath / prediction["pk"]
        for socket_value in prediction["outputs"]:
            generate_socket_value_stub_file(
                output_zip_file=output_zip_file,
                target_zpath=prediction_zpath
                / "output"
                / socket_value["interface"]["relative_path"],
                socket=socket_value["interface"],
            )


def generate_algorithm_template(
    *,
    context,
    output_zip_file,
    target_zpath,
):
    validate_algorithm_template_context(context)

    context["grand_challenge_forge_version"] = metadata.version(
        "grand-challenge-forge"
    )

    generate_example_algorithm(
        context={"phase": context["algorithm"]},
        output_zip_file=output_zip_file,
        target_zpath=target_zpath,
    )

    copy_and_render(
        templates_dir_name="algorithm-template-readme",
        output_zip_file=output_zip_file,
        target_zpath=target_zpath,
        context=context,
    )



import json
import logging
import os
import time
import uuid
import zipfile
from datetime import datetime, timezone
from pathlib import Path

import black
from jinja2 import FileSystemLoader, StrictUndefined, TemplateNotFound
from jinja2.sandbox import ImmutableSandboxedEnvironment

from grand_challenge_forge import PARTIALS_PATH

DEBUG = os.getenv("GRAND_CHALLENGE_FORGE_DEBUG", "false").lower() == "true"

SCRIPT_PATH = Path(os.path.dirname(os.path.realpath(__file__)))
RESOURCES_PATH = SCRIPT_PATH / "resources"


logger = logging.getLogger(__name__)


def is_json(socket):
    return socket["relative_path"].endswith(".json")


def is_image(socket):
    return socket["super_kind"] == "Image"


def is_file(socket):
    return socket["super_kind"] == "File" and not socket[
        "relative_path"
    ].endswith(".json")


def has_example_value(socket):
    return "example_value" in socket and socket["example_value"] is not None


def generate_socket_value_stub_file(*, output_zip_file, target_zpath, socket):
    """Creates a stub based on a component interface"""
    if has_example_value(socket):
        zinfo = zipfile.ZipInfo(str(target_zpath))
        output_zip_file.writestr(
            zinfo,
            json.dumps(
                socket["example_value"],
                indent=4,
            ),
        )
        return

    # Copy over an example

    if is_json(socket):
        source = RESOURCES_PATH / "example.json"
    elif is_image(socket):
        source = RESOURCES_PATH / "example.mha"
        target_zpath = target_zpath / f"{str(uuid.uuid4())}.mha"
    else:
        source = RESOURCES_PATH / "example.txt"

    output_zip_file.write(
        source,
        arcname=str(target_zpath),
    )


def socket_to_socket_value(socket):
    """Creates a stub dict repr of a socket valuee"""
    sv = {
        "file": None,
        "image": None,
        "value": None,
    }
    if socket["super_kind"] == "Image":
        sv["image"] = {
            "name": "the_original_filename_of_the_file_that_was_uploaded.suffix",
        }
    if socket["super_kind"] == "File":
        sv["file"] = (
            f"https://grand-challenge.org/media/some-link/"
            f"{socket['relative_path']}"
        )
    if socket["super_kind"] == "Value":
        sv["value"] = socket.get("example_value", {"some_key": "some_value"})
    return {
        **sv,
        "interface": socket,
    }


def get_jinja2_environment(searchpath=None):
    from grand_challenge_forge.partials.filters import custom_filters

    if searchpath:
        searchpath = [searchpath, PARTIALS_PATH]
    else:
        searchpath = PARTIALS_PATH

    env = ImmutableSandboxedEnvironment(
        loader=FileSystemLoader(
            searchpath=searchpath,
            followlinks=True,
        ),
        undefined=StrictUndefined,
        keep_trailing_newline=True,
    )
    env.filters = custom_filters
    env.filters["zip"] = zip
    env.globals["now"] = datetime.now(timezone.utc)

    return env


def copy_and_render(
    *,
    templates_dir_name,
    output_zip_file,
    target_zpath,
    context,
):
    source_path = PARTIALS_PATH / templates_dir_name

    if not source_path.exists():
        raise TemplateNotFound(source_path)

    env = get_jinja2_environment(searchpath=source_path)

    for root, _, files in os.walk(source_path, followlinks=True):
        root = Path(root)

        check_allowed_source(path=root)

        # Create relative path
        rel_path = root.relative_to(source_path)
        current_zdir = target_zpath / rel_path

        for file in sorted(files):
            source_file = root / file
            output_file = current_zdir / file

            check_allowed_source(path=source_file)

            if file.endswith(".j2"):  # Jinja2 template
                template = env.get_template(
                    name=str(source_file.relative_to(source_path))
                )
                rendered_content = template.render(
                    **context,
                    _no_gpus=DEBUG,
                )

                targetfile_zpath = output_file.with_suffix("")

                if targetfile_zpath.suffix == ".py":
                    rendered_content = apply_black(rendered_content)

                # Collect information about the file to be written to the zip file
                # (permissions, et cetera)
                zinfo = zipfile.ZipInfo.from_file(
                    source_file,
                    arcname=str(targetfile_zpath),
                )

                # Update the date time of creation, since we are technically
                # creating a new file
                # Also (partially) addresses a problem where docker build injects
                # incorrect files:
                # https://github.com/moby/buildkit/issues/4817#issuecomment-2032551066
                zinfo.date_time = time.localtime()[0:6]

                output_zip_file.writestr(zinfo, rendered_content)
            else:
                output_zip_file.write(
                    str(source_file), arcname=str(output_file)
                )


def check_allowed_source(path):
    if PARTIALS_PATH.resolve() not in path.resolve().parents:
        raise PermissionError(
            f"Only files under {PARTIALS_PATH} are allowed "
            "to be copied or rendered"
        )


def apply_black(content):
    # Format rendered Python code string using black
    result = black.format_str(
        content,
        mode=black.Mode(),
    )
    return result



import json
import logging
import shutil
from importlib import metadata
from pathlib import Path

import click

from grand_challenge_forge import logger
from grand_challenge_forge.exceptions import ChallengeForgeError
from grand_challenge_forge.forge import (
    generate_algorithm_template,
    generate_challenge_pack,
)
from grand_challenge_forge.utils import truncate_with_epsilons
from tests.utils import zipfile_to_filesystem


# Shared options decorator
def common_options(func):
    """Decorator to add common options to multiple commands."""
    func = click.option(
        "-o",
        "--output",
        type=click.Path(
            exists=False,
            file_okay=False,
            dir_okay=True,
            readable=True,
            writable=True,
            resolve_path=True,
        ),
        default="dist/",
        show_default=True,
    )(func)
    func = click.option(
        "-f",
        "--force",
        is_flag=True,
        default=False,
    )(func)
    func = click.argument(
        "contexts",
        nargs=-1,
    )(func)
    func = click.option(
        "-v",
        "--verbose",
        count=True,
        help="Sets verbosity level. Stacks (e.g. -vv = debug)",
    )(func)
    return func


@click.group()
@click.version_option(metadata.version("grand-challenge-forge"), "--version")
def cli():
    """Main CLI entry point."""
    pass


@cli.command()
@common_options
def pack(output, force, contexts, verbose=0):
    """
    Generates a challenge pack using provided context.

    A context can be a filename or a JSON string.

    Multiple contexts can be provided. Each will be processed independently.
    """
    output_dir = Path(output)

    _set_verbosity(verbosity=verbose)

    for index, context in enumerate(contexts):
        resolved_context = _resolve_context(src=context)
        if resolved_context:
            try:
                logger.info(
                    f"🏗️Started working on pack [{index + 1} of {len(contexts)}]"
                )
                with zipfile_to_filesystem(output_path=output_dir) as zip_file:
                    pack_zpath = generate_challenge_pack(
                        target_zpath=Path(
                            f"{resolved_context['challenge']['slug']}-challenge-pack"
                        ),
                        context=resolved_context,
                        output_zip_file=zip_file,
                    )
                    pack_dir = output_dir / pack_zpath

                    if pack_dir.exists():
                        if force:
                            shutil.rmtree(pack_dir)
                        else:
                            raise ChallengeForgeError(
                                f"Pack {pack_dir.stem!r} already exists!"
                            )

                logger.info(f"📦 Created Pack {pack_dir.stem!r}")
                logger.info(f"📢 Pack is here: {pack_dir}")
                print(str(pack_dir))
            except Exception as e:
                if isinstance(e, ChallengeForgeError):
                    logger.error(f"💔 {e}", exc_info=True)
                else:
                    raise e


@cli.command()
@common_options
def algorithm(output, force, contexts, verbose):
    """
    Generates an algorithm template using provided context.

    A context can be a filename or a JSON string.

    Multiple contexts can be provided. Each will be processed independently.
    """

    output_dir = Path(output)

    _set_verbosity(verbosity=verbose)

    for index, context in enumerate(contexts):
        resolved_context = _resolve_context(src=context)
        if resolved_context:
            try:
                logger.info(
                    f"🏗️Started working on Algorithm Template [{index + 1} "
                    f"of {len(contexts)}]"
                )

                with zipfile_to_filesystem(output_path=output_dir) as zip_file:
                    template_zpath = generate_algorithm_template(
                        target_zpath=Path(
                            f"{resolved_context['algorithm']['slug']}-template"
                        ),
                        context=resolved_context,
                        output_zip_file=zip_file,
                    )
                    template_dir = output_dir / template_zpath

                    if template_dir.exists():
                        if force:
                            shutil.rmtree(template_dir)
                        else:
                            raise ChallengeForgeError(
                                f"Algorithm Template {template_dir.stem!r} "
                                "already exists!"
                            )
                logger.info(
                    f"📦 Created Algorithm Template {template_dir.stem!r}"
                )
                logger.info(f"📢 Algorithm Template is here: {template_dir}")
                print(str(template_dir))
            except Exception as e:
                if isinstance(e, ChallengeForgeError):
                    logger.error(f"💔 {e}")
                else:
                    raise


def _set_verbosity(verbosity):
    ch = logging.StreamHandler()

    if verbosity == 0:
        logger.setLevel(logging.WARNING)
        ch.setLevel(logging.WARNING)
    elif verbosity == 1:
        logger.setLevel(logging.INFO)
        ch.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.DEBUG)
        ch.setLevel(logging.DEBUG)


def _resolve_context(src):
    try:
        if (p := Path(src)).exists() and p.is_file():
            return _read_json_file(p)
        return json.loads(src)
    except json.decoder.JSONDecodeError as e:
        logger.error(
            f"Could not resolve context source:\n"
            f"'{truncate_with_epsilons(src)!r}' {e}"
        )


def _read_json_file(json_file):
    with open(json_file, "r") as f:
        context = json.load(f)
    return context



import logging
import os
from pathlib import Path

logging.basicConfig(format="%(levelname)s: %(message)s")
logger = logging.getLogger(__name__)
SCRIPT_PATH = Path(os.path.dirname(os.path.realpath(__file__)))
PARTIALS_PATH = SCRIPT_PATH / "partials"
RESOURCES_PATH = SCRIPT_PATH / "resources"



from grand_challenge_forge import generation_utils

custom_filters = {}


def register_simple_filter(func):
    custom_filters[func.__name__] = func
    return func


@register_simple_filter
def is_json(arg):
    return generation_utils.is_json(arg)


@register_simple_filter
def has_json(arg):
    return any(generation_utils.is_json(item) for item in arg)


@register_simple_filter
def is_image(arg):
    return generation_utils.is_image(arg)


@register_simple_filter
def has_image(arg):
    return any(generation_utils.is_image(item) for item in arg)


@register_simple_filter
def is_file(arg):
    return generation_utils.is_file(arg)


@register_simple_filter
def has_file(arg):
    return any(generation_utils.is_file(item) for item in arg)


@register_simple_filter
def has_example_value(arg):
    return generation_utils.has_example_value(arg)
